import secrets
from datetime import datetime, timedelta, timezone
from .states import AlarmState, AlarmAttrs
from .trigger import Trigger, TriggerType
from ..tags.tag import Tag, MachineObserver
from ..tags.cvt import CVTEngine
from ..modules.users.users import User
from ..utils.decorators import validate_types, logging_error_handler, set_event, put_alarm_state
from ..models import FloatType, IntegerType, StringType
from ..variables import *
from statemachine import State, StateMachine
from flask_socketio import SocketIO


class Alarm(StateMachine):
    r"""
    Represents an Alarm entity with state machine logic.

    Implements the standard alarm lifecycle defined in ISA 18.2, including states like
    Normal, Unacknowledged, Acknowledged, Shelved, and Suppressed.

    It monitors a `Tag` value against a `Trigger` condition and transitions states accordingly.
    """

    # MAIN STATES
    normal = State("normal", initial=True)
    unack_alarm = State("unack_alarm")
    ack_alarm = State("ack_alarm")
    rtn_unack = State("rtn_unack")

    # OUT OF SERVICE STATES
    shelved = State("shelved")
    suppressed_by_design = State("suppressed_by_design")
    out_of_service = State("out_of_service")

    # MAIN TRANSITIONS
    normal_to_unack_alarm = normal.to(unack_alarm)
    unack_alarm_to_ack_alarm = unack_alarm.to(ack_alarm)
    ack_alarm_to_normal = ack_alarm.to(normal)
    unack_alarm_to_rtn_unack = unack_alarm.to(rtn_unack)
    rtn_unack_to_normal = rtn_unack.to(normal)
    rtn_unack_to_unack_alarm = rtn_unack.to(unack_alarm)

    # SHELVED TRANSITIONS
    normal_to_shelved = normal.to(shelved)
    unack_alarm_to_shelved = unack_alarm.to(shelved)
    ack_alarm_to_shelved = ack_alarm.to(shelved)
    rtn_unack_to_shelved = rtn_unack.to(shelved)
    shelved_to_normal = shelved.to(normal)
    shelved_to_unack_alarm = shelved.to(unack_alarm)

    # SUPPRESSED BY DESIGN TRANSITIONS
    normal_to_suppressed_by_design = normal.to(suppressed_by_design)
    unack_alarm_to_suppressed_by_design = unack_alarm.to(suppressed_by_design)
    ack_alarm_to_suppressed_by_design = ack_alarm.to(suppressed_by_design)
    rtn_unack_to_suppressed_by_design = rtn_unack.to(suppressed_by_design)
    suppressed_by_design_to_normal = suppressed_by_design.to(normal)
    suppressed_by_design_to_unack_alarm = suppressed_by_design.to(unack_alarm)


    # OUT OF SERVICE TRANSITIONS
    normal_to_out_of_service = normal.to(out_of_service)
    unack_alarm_to_out_of_service = unack_alarm.to(out_of_service)
    ack_alarm_to_out_of_service = ack_alarm.to(out_of_service)
    rtn_unack_to_out_of_service = rtn_unack.to(out_of_service)
    out_of_service_to_normal = out_of_service.to(normal)
    out_of_service_to_unack_alarm = out_of_service.to(unack_alarm)

    def __init__(
            self,
            name:str, 
            tag:Tag,
            alarm_type:StringType,
            alarm_setpoint:IntegerType|FloatType,
            description:str="",
            state:str=None,
            timestamp:datetime=None,
            ack_timestamp:datetime=None,
            alarm_deadband:IntegerType|FloatType=FloatType(0.0),
            alarm_on_delay:IntegerType|FloatType=FloatType(0.0),
            alarm_off_delay:IntegerType|FloatType=FloatType(0.0),
            identifier:str=None,
            user:User=None,
            reload:bool=False
        ):
        r"""
        Initializes the Alarm.

        **Parameters:**

        * **name** (str): Alarm name.
        * **tag** (Tag): The tag being monitored.
        * **alarm_type** (StringType): Trigger type (HI, LO, etc.).
        * **alarm_setpoint** (IntegerType|FloatType): The limit value.
        * **description** (str): Alarm description.
        * **state** (str, optional): Initial state.
        * **identifier** (str, optional): Unique ID.
        """
        from ..logger.alarms import AlarmsLoggerEngine
        self.alarm_engine = AlarmsLoggerEngine()
        self.tag_engine = CVTEngine()
        self.name = name
        self.tag = tag
        self.segment = tag.segment
        self.manufacturer = tag.manufacturer
        self.attach(machine=self, tag=tag)
        self.description = description        
        self.alarm_setpoint = Trigger()
        self.alarm_setpoint.value = alarm_setpoint.value
        self.alarm_setpoint.type = TriggerType(value=alarm_type.value.upper())
        alarm_deadband.unit = tag.get_display_unit()
        self.alarm_deadband = alarm_deadband
        self.alarm_on_delay = alarm_on_delay
        self.alarm_off_delay = alarm_off_delay
        self.timestamp = timestamp 
        self.ack_timestamp = ack_timestamp
        self.state = AlarmState.NORM
        if state:
            
            for _, attr in AlarmState.__dict__.items():
                
                if isinstance(attr, AlarmAttrs):
                    
                    if state==attr.state:
                        
                        self.state = attr
                        break
        if identifier:
            self.identifier = identifier
        else:
            self.identifier = secrets.token_hex(4)

        self._shelved_time:datetime = None
        self._shelved_until:datetime = None
        self._shelved_options_time = {
            'days': 0,
            'seconds': 0,
            'microseconds': 0,
            'milliseconds': 0,
            'minutes': 0,
            'hours': 0,
            'weeks': 0
        }
        transitions = []
        for state in self.states:
            transitions.extend(state.transitions)
        self.transitions = transitions
        self.sio:SocketIO|None = None
        super(Alarm, self).__init__()

    @logging_error_handler
    @put_alarm_state
    def on_enter_normal(self):
        self.state = AlarmState.NORM
        self.timestamp = None 
        self.ack_timestamp = None

    @logging_error_handler
    @put_alarm_state
    def on_enter_unack_alarm(self):

        self.state = AlarmState.UNACK
        self.timestamp = self.__timestamp
        self.alarm_engine.create_record_on_alarm_summary(
                name=self.name, 
                state=self.state.state, 
                timestamp=self.timestamp,
                ack_timestamp=self.ack_timestamp
            )

    @logging_error_handler
    @put_alarm_state
    def on_enter_ack_alarm(self):

        self.state = AlarmState.ACKED
        self.ack_timestamp = self.__timestamp
        self.alarm_engine.put_record_on_alarm_summary(
            name=self.name, 
            state=self.state.state, 
            ack_timestamp=self.ack_timestamp
        )
    
    @logging_error_handler
    @put_alarm_state
    def on_enter_rtn_unack(self):
        
        self.timestamp = None
        self.state = AlarmState.RTNUN
        self.alarm_engine.put_record_on_alarm_summary(
            name=self.name, 
            state=self.state.state
        )
    
    @logging_error_handler
    @put_alarm_state
    def on_enter_shelved(self):
        
        self.state = AlarmState.SHLVD

    @logging_error_handler
    @put_alarm_state
    def on_enter_suppressed_by_design(self):
        
        self.state = AlarmState.DSUPR

    @logging_error_handler
    @put_alarm_state
    def on_enter_out_of_service(self):
        
        self.state = AlarmState.OOSRV

    def set_socketio(self, sio:SocketIO):
        r"""
        Sets the SocketIO instance for real-time updates.

        **Parameters:**

        * **sio** (SocketIO): The SocketIO server instance.
        """
        self.sio:SocketIO = sio

    @logging_error_handler
    @validate_types(
            tag=str, 
            value=Temperature|Length|Current|Time|Pressure|Mass|Force|Power|VolumetricFlow|MassFlow|Density|Percentage|Adimentional, 
            timestamp=datetime, 
            output=None)
    def notify(
        self, 
        tag:str, 
        value:Temperature|Length|Current|Time|Pressure|Mass|Force|Power|VolumetricFlow|MassFlow|Density|Percentage|Adimentional, 
        timestamp:datetime):
        r"""
        Callback triggered when the monitored Tag value changes (Observer pattern).
        
        It evaluates the new value against alarm logic and triggers state transitions.

        **Parameters:**

        * **tag** (str): Tag name.
        * **value** (Quantity): The new tag value.
        * **timestamp** (datetime): Time of the value change.
        """ 
        self.__timestamp = timestamp
        if self.state not in (AlarmState.DSUPR, AlarmState.SHLVD, AlarmState.OOSRV):     
            if self.alarm_setpoint.type in (TriggerType.HH, TriggerType.H):

                if value.value > self.alarm_setpoint.value:

                    self.abnormal_condition()
                
                else: 
                    self.normal_condition()

            elif self.alarm_setpoint.type in (TriggerType.L, TriggerType.LL):

                if value.value < self.alarm_setpoint.value:

                    self.abnormal_condition()

                else:

                    self.normal_condition()

            else: # Boolean Alarm
                
                if value.value == bool(self.alarm_setpoint.value):

                    self.abnormal_condition()

                else:

                    self.normal_condition()

        if self.state==AlarmState.SHLVD:

            if datetime.now(timezone.utc) >= self._shelved_until:

                self.unshelve()
        
    @logging_error_handler
    def abnormal_condition(self):
        r"""
        Triggers transition to an alarm state (abnormal).
        """
        current_state = self.current_state.name.lower()
        transition_name = f'{current_state}_to_unack_alarm'
        self.__transition(transition_name=transition_name)

    @logging_error_handler
    def normal_condition(self):
        r"""
        Triggers transition to normal or return-to-normal states.
        """
        current_state = self.current_state.name.lower()

        if current_state=="unack_alarm":

            transition_name = f'{current_state}_to_rtn_unack'
            self.__transition(transition_name=transition_name)

        elif current_state=="ack_alarm":

            transition_name = f'{current_state}_to_normal'
            self.__transition(transition_name=transition_name)

    @logging_error_handler
    @set_event(message=f"Acknowledged", classification="Alarm", priority=2, criticity=3)
    def acknowledge(self, user:User=None):
        r"""
        Acknowledges the alarm.

        **Parameters:**

        * **user** (User, optional): User performing the acknowledgment.
        """
        current_state = self.current_state.name.lower()

        if current_state=="unack_alarm":

            transition_name = f'{current_state}_to_ack_alarm'

        elif current_state=="rtn_unack":

            transition_name = f'{current_state}_to_normal'


        tag = self.tag_engine.get_tag_by_name(name=self.tag.name)
        self.alarm_engine.put_record_on_alarm_summary(
            name=self.name, 
            ack_timestamp=tag.get_timestamp()
        )
        self.__transition(transition_name=transition_name)
        return self, f"{self.tag.get_name()}"

    @logging_error_handler
    @set_event(message=f"Shelved", classification="Alarm", priority=2, criticity=3)
    def shelve(self, user:User=None, **options):
        r"""
        Temporarily suppresses the alarm (Shelving).

        **Parameters:**

        * **options**: Time duration arguments (days, hours, minutes, seconds).
        """
        options_time = {key: options[key] if key in options else self._shelved_options_time[key] for key in self._shelved_options_time}
        
        if options_time!=self._shelved_options_time:
            
            self._shelved_time = datetime.now(timezone.utc)
            self._shelved_until = self._shelved_time + timedelta(**options_time)

        current_state = self.current_state.name.lower()
        transition_name = f'{current_state}_to_shelved'
        self.__transition(transition_name=transition_name)
        return self, f"{self.tag.get_name()}"

    @logging_error_handler
    @set_event(message=f"Unshelved", classification="Alarm", priority=2, criticity=3)
    def unshelve(self, user:User=None):
        r"""
        Manually un-shelves the alarm, returning it to service.
        """
        self.__return_to_service()
        return self, f"{self.tag.get_name()}"

    @logging_error_handler
    @set_event(message=f"Designed suppression", classification="Alarm", priority=2, criticity=3)
    def designed_suppression(self, user:User=None):
        r"""
        Suppresses the alarm by design (e.g., maintenance mode).
        """
        current_state = self.current_state.name.lower()
        transition_name = f'{current_state}_to_suppressed_by_design'
        self.__transition(transition_name=transition_name)
        return self, f"{self.tag.get_name()}"

    @logging_error_handler
    @set_event(message=f"Designed unsuppression", classification="Alarm", priority=2, criticity=3)
    def designed_unsuppression(self, user:User=None):
        r"""
        Removes designed suppression.
        """
        self.__return_to_service()
        return self, f"{self.tag.get_name()}"

    @logging_error_handler
    @set_event(message=f"Removed from service", classification="Alarm", priority=2, criticity=3)
    def remove_from_service(self, user:User=None):
        r"""
        Takes the alarm out of service entirely.
        """
        current_state = self.current_state.name.lower()
        transition_name = f'{current_state}_to_out_of_service'
        self.__transition(transition_name=transition_name)
        return self, f"{self.tag.get_name()}"

    @logging_error_handler
    @set_event(message=f"Returned to service", classification="Alarm", priority=2, criticity=3)
    def return_to_service(self, user:User=None):
        r"""
        Returns the alarm to service from 'Out of Service'.
        """
        self.__return_to_service()

        return self, f"{self.tag.get_name()}"

    @logging_error_handler
    def attach(self, machine, tag:Tag):
        
        def attach_observer(machine, tag:Tag):

            observer = MachineObserver(machine)
            query = dict()
            query["action"] = "attach_observer"
            query["parameters"] = {
                "name": tag.name,
                "observer": observer,
            }
            self.tag_engine.request(query)
            self.tag_engine.response()

        attach_observer(machine, tag)

    @set_event(message=f"Updated", classification="Alarm", priority=2, criticity=3)
    def put(
            self, 
            user:User=None,
            name:str=None,
            tag:str=None,
            description:str=None,
            alarm_type:TriggerType=None,
            trigger_value:float=None):
        r"""
        Updates the alarm configuration.

        **Parameters:**
        
        * **name** (str): Alarm name.
        * **tag** (str): Tag bound to alarm.
        * **description** (str): Alarm description.
        * **alarm_type** (TriggerType): Alarm type ['HIGH-HIGH', 'HIGH', 'LOW', 'LOW-LOW', 'BOOL'].
        * **trigger_value** (float): Alarm trigger value.

        **Returns:**

        * **tuple**: (Alarm instance, status message)
        """
        message = ""
        if alarm_type:

            if alarm_type.value.upper() in ["HIGH-HIGH", "HIGH", "LOW", "LOW-LOW", "BOOL"]:

                self.alarm_setpoint.type = alarm_type

                message += f" alarm_type: {alarm_type.value}"

        if trigger_value:
            
            self.alarm_setpoint.value = float(trigger_value)
            message += f" trigger value: {trigger_value}"
        
        if name:

            self._name = name
            message += f" name: {name}"

        if tag:

            self._tag = tag
            message += f" tag: {tag}"

        if description:

            self._description = description
            message += f" description: {description}"
        
        return self, message

    def _get_active_transitions(self):
        r"""
        Gets allowed transitions based on the current state.

        **Returns:**

        * **list**: List of available transitions.
        """
        result = list()

        current_state = self.current_state
        transitions = self.transitions

        for transition in transitions:

            if transition.source == current_state:

                result.append(transition)

        return result
    
    @logging_error_handler
    def __transition(self, transition_name:str):

        allowed_transitions = self._get_active_transitions()
        for _transition in allowed_transitions:
            
            if f"{_transition.source.name}_to_{_transition.target.name}"==transition_name:
                
                self.send(transition_name)

    @logging_error_handler
    def __return_to_service(self):

        current_state = self.current_state.name.lower()

        if self.state.alarm_status.lower()=="active":

            transition_name = f'{current_state}_to_unack_alarm'

        else:

            transition_name = f'{current_state}_to_normal'
        
        self.__transition(transition_name=transition_name)

    @logging_error_handler
    def get_operator_actions(self)->list:
        r"""
        Returns a list of available actions for the operator based on current state.

        **Returns:**

        * **dict**: Map of Action Name -> Action Method.
        """
        current_state = self.current_state.name.lower()
            
        if current_state in ("unack_alarm", "rtn_unack"):

            result = {
                "Acknowledge": "acknowledge",
                "Shelve": "shelve",
                "Designed Suppression": "designed_suppression",
                "Remove From Service": "remove_from_service"
            }

        elif current_state=="suppressed_by_design":

            result = {
                "Designed Unsuppression": "designed_unsuppression"
            }

        elif current_state=="out_of_service":

            result = {
                "Return To Service": "return_to_service"
            }

        else:

            result = {
                "Shelve": "shelve",
                "Designed Suppression": "designed_suppression",
                "Remove From Service": "remove_from_service"
            }

        return result

    def serialize(self):
        r"""
        Serializes the alarm object to a JSON-compatible dictionary.

        **Returns:**

        * **dict**: Alarm data including state, setpoint, and metadata.
        """
        timestamp = self.timestamp
        if timestamp:

            timestamp = timestamp.strftime(self.tag_engine.DATETIME_FORMAT)

        ack_timestamp = self.ack_timestamp
        if ack_timestamp:

            ack_timestamp = ack_timestamp.strftime(self.tag_engine.DATETIME_FORMAT)

        return {
            "identifier": self.identifier,
            "segment": self.segment,
            "manufacturer": self.manufacturer,
            "timestamp": timestamp,
            "name": self.name,
            "tag": self.tag.name,
            "state": self.state.serialize(),
            "alarm_setpoint": self.alarm_setpoint.serialize(),
            "ack_timestamp": ack_timestamp,
            "description": self.description,
            "actions": self.get_operator_actions()
        }



from peewee import CharField, IntegerField
from ..dbmodels.core import BaseModel

class OPCUA(BaseModel):
    r"""
    Database model for OPC UA Client configurations.
    """
    
    client_name = CharField(unique=True)
    host = CharField()
    port = IntegerField()

    @classmethod
    def create(cls, client_name:str, host:str, port:int):
        r"""
        Creates a new OPC UA client configuration.

        **Parameters:**

        * **client_name** (str): Unique name for the client.
        * **host** (str): Server hostname or IP.
        * **port** (int): Server port.

        **Returns:**

        * **OPCUA**: The created record.
        """

        if not cls.client_name_exist(client_name):

            query = cls(client_name=client_name, host=host, port=port)
            query.save()

            return query
        
    @classmethod
    def get_by_client_name(cls, client_name:str):
        r"""
        Retrieves a client configuration by name.
        """
        return cls.get_or_none(client_name=client_name)
    
    @classmethod
    def client_name_exist(cls, client_name:str):
        r"""
        Checks if a client name exists.
        """
        query = cls.get_or_none(client_name=client_name)
        
        if query is not None:

            return True
        
        return False
    
    def serialize(self):
        r"""
        Serializes the client configuration.
        """
        return {
            "client_name": self.client_name,
            "host": self.host,
            "port": self.port
        }


import dash
from ...utils import find_differences_between_lists, generate_dropdown_conditional
from ...variables import VARIABLES

def init_callback(app:dash.Dash):

    @app.callback(
        dash.Input("tag_name_input", "value"),
        dash.Input("variable_input", "value"),
        dash.Input("datatype_input", "value"),
        dash.Input("unit_input", "value"),  
        dash.Input("display_unit_input", "value"), 
        dash.Input("manufacturer_input", "value"), 
        dash.Input("segment_input", "value"),
        dash.Input("segment_radio_button", "value")
        )
    def create_tag(
        name:str, 
        variable:str, 
        datatype:str, 
        unit:str, 
        display_unit:str,
        manufacturer:str,
        segment:str,
        segment_bool:bool
        )->str:
        r"""
        Documentation here
        """
        if dash.ctx.triggered_id.lower()=="variable_input":

            options = [{"label": value, "value": value} for _, value in VARIABLES[variable].items()]
            unit = list(VARIABLES[variable].values())[0]
            
            dash.set_props("unit_input", {'options': options})
            dash.set_props("display_unit_input", {'options': options})
            dash.set_props("unit_input", {'value': unit})
            dash.set_props("unit_input", {'disabled': False})
            dash.set_props("display_unit_input", {'value': unit})
            dash.set_props("display_unit_input", {'disabled': False})

        if not segment_bool:

            if name and datatype and unit and variable and display_unit:

                dash.set_props("create_tag_button", {'disabled': False})

            else:
                
                dash.set_props("create_tag_button", {'disabled': True})

        else:

            if name and datatype and unit and variable and display_unit and manufacturer and segment:

                dash.set_props("create_tag_button", {'disabled': False})

            else:
                
                dash.set_props("create_tag_button", {'disabled': True})

    @app.callback(
        dash.Output("description_input", "value"),
        dash.Input("description_radio_button", "value")
    )
    def enable_description(enable:bool):
        r"""
        Documentation here
        """
        dash.set_props("description_input", {'disabled': not enable})
        return ""
    
    @app.callback(
        dash.Output("segment_input", "value"),
        dash.Output("manufacturer_input", "value", allow_duplicate=True),
        dash.Input("segment_radio_button", "value")
    )
    def enable_segment(enable:bool):
        r"""
        Documentation here
        """
        dash.set_props("segment_input", {'disabled': not enable})
        dash.set_props("manufacturer_input", {'disabled': not enable})
        return "", ""
    
    @app.callback(
        dash.Output("node_namespace_input", "value"),
        dash.Input("opcua_address_input", "value")
    )
    def enable_node_namespace(opcua_server:str):
        r"""
        Documentation here
        """
        if opcua_server:
            
            dash.set_props("node_namespace_input", {'disabled': False})

        else:
            
            dash.set_props("node_namespace_input", {'disabled': True})

        return ""
    
    @app.callback(
        dash.Output('scan_time_input', 'value'),
        dash.Output('dead_band_input', 'value'),
        dash.Input("node_namespace_input", "value")
    )
    def enable_scan_time_and_dead_band(node_namespace:str):
        r"""
        Documentation here
        """
        if node_namespace:

            dash.set_props("scan_time_input", {'disabled': False})
            dash.set_props("dead_band_input", {'disabled': False})

        else:

            dash.set_props("scan_time_input", {'disabled': True})
            dash.set_props("dead_band_input", {'disabled': True})

        return "", ""
    
    @app.callback(
        dash.Output("display_name_input", "value"),
        dash.Input("display_name_radio_button", "value")
    )
    def enable_display(enable:bool):
        r"""
        Documentation here
        """
        dash.set_props("display_name_input", {'disabled': not enable})
        return ""
    
    @app.callback(
        dash.Output("dead_band_unit", "children"),
        dash.Input("unit_input", "value")
    )
    def update_unit(unit:str):
        r"""
        Documentation here
        """
        return unit
    
    @app.callback(
        dash.Output('tags_datatable', 'data', allow_duplicate=True),
        dash.Output('opcua_address_input', 'options'),
        dash.Output('tags_datatable', 'dropdown', allow_duplicate=True),
        dash.Output('tags_datatable', 'dropdown_conditional', allow_duplicate=True),
        dash.Input('tags_page', 'pathname'),
        prevent_initial_call=True
        )
    def display_page(pathname):
        r"""
        Documentation here
        """
        
        opcua_client_options = [{"label": "", "value": ""}]

        for opcua_client, info in app.automation.get_opcua_clients().items():
            
            opcua_client_options.append({
                "label": opcua_client, "value": info['server_url']
            })
        
        dropdown = {
            'data_type': {
                'options': [
                    {'label': 'Float', 'value': 'float'},
                    {'label': 'Integer', 'value': 'integer'},
                    {'label': 'Boolean', 'value': 'boolean'},
                    {'label': 'String', 'value': 'string'}
                ],
                'clearable': False
            },
            'opcua_address': {
                'options': opcua_client_options
            },
            'segment': {
                'options': [{'label': f"{segment['manufacturer']['name']}->{segment['name']}", 'value': f"{segment['manufacturer']['name']}->{segment['name']}"} for segment in app.automation.get_segments() if app.automation.get_segments()]
            },
            'variable': {
                'options': [{"label": key, "value": key} for key, _ in VARIABLES.items()]
            }
        }
        dropdown_conditional = generate_dropdown_conditional()
        if pathname=="/tags":
            return app.tags_table_data(), opcua_client_options, dropdown, dropdown_conditional
        
        return dash.no_update, opcua_client_options, dropdown, dropdown_conditional
        
    @app.callback(
        dash.Output('node_namespace_input', 'options'),
        dash.Input('opcua_address_input', 'value'),
        prevent_initial_call=True
        )
    def select_opcua_server(server_url):
        r"""
        Documentation here
        """
        opcua_clients = app.automation.get_opcua_clients()
        for client_name, info in opcua_clients.items():

            if server_url==info["server_url"]:

                break

        nodes = [{"label": "None", "value": "None"}]
        tree = app.automation.get_opcua_tree(client_name)
        for node in tree[0]["Objects"][0]["children"]:

            nodes.append(
                {
                    "label": node["title"],
                    "value": node["key"]
                }
            )
        return nodes
        
    @app.callback(
        dash.Output('tags_datatable', 'data', allow_duplicate=True),
        dash.Output('tags_datatable', 'dropdown_conditional', allow_duplicate=True),
        dash.Output('tags_datatable', 'dropdown', allow_duplicate=True),
        dash.Input('create_tag_button', 'n_clicks'),
        dash.State("tag_name_input", "value"), 
        dash.State("datatype_input", "value"), 
        dash.State("unit_input", "value"),
        dash.State("display_unit_input", "value"), 
        dash.State("variable_input", "value"),
        dash.State("display_name_input", "value"), 
        dash.State("description_input", "value"),
        dash.State("manufacturer_input", "value"),
        dash.State("segment_input", "value"),
        dash.State("opcua_address_input", "value"),
        dash.State("node_namespace_input", "value"),
        dash.State("scan_time_input", "value"),
        dash.State("dead_band_input", "value"),
        prevent_initial_call=True
    )
    def displayClick(
        btn1, 
        tag_name,
        datatype,
        unit,
        display_unit,
        variable,
        display_name,
        description,
        manufacturer,
        segment,
        opcua_address,
        node_namespace,
        scan_time:int|None=None,
        dead_band:float|None=None
        ):
        r"""
        Documentation here
        """
        if "create_tag_button" == dash.ctx.triggered_id:

            if not scan_time:

                scan_time = None
            
            if not dead_band:

                dead_band = None
            
            tag, message = app.automation.create_tag(
                name=tag_name,
                unit=unit,
                display_unit=display_unit,
                data_type=datatype,
                variable=variable,
                description=description,
                display_name=display_name,
                opcua_address=opcua_address,
                node_namespace=node_namespace,
                scan_time=scan_time,
                dead_band=dead_band,
                manufacturer=manufacturer,
                segment=segment
            )
            if not tag:
                segment = {
                    'options': []
                }
                dash.set_props("modal-body", {"children": message})
                dash.set_props("modal-centered", {'is_open': True})
            
            else:
                segment = {
                    'options': [{'label': f"{segment['manufacturer']['name']}->{segment['name']}", 'value': f"{segment['manufacturer']['name']}->{segment['name']}"} for segment in app.automation.get_segments() if app.automation.get_segments()]
                }
                dash.set_props("modal-success-body", {"children": message})
                dash.set_props("modal-success", {'is_open': True})

            opcua_client_options = [{"label": "", "value": ""}]

            for opcua_client, info in app.automation.get_opcua_clients().items():
                
                opcua_client_options.append({
                    "label": opcua_client, "value": info['server_url']
                })
            dropdown = {
                'data_type': {
                    'options': [
                        {'label': 'Float', 'value': 'float'},
                        {'label': 'Integer', 'value': 'integer'},
                        {'label': 'Boolean', 'value': 'boolean'},
                        {'label': 'String', 'value': 'string'}
                    ],
                    'clearable': False
                },
                'opcua_address': {
                    'options': opcua_client_options
                },
                'segment': segment,
                'variable': {
                    'options': [{"label": key, "value": key} for key, _ in VARIABLES.items()]
                }
            }
        
            return app.tags_table_data(), generate_dropdown_conditional(), dropdown
        
    @app.callback(
        dash.Input('tags_datatable', 'data_timestamp'),
        dash.State('tags_datatable', 'data_previous'),
        dash.State('tags_datatable', 'data'),
        )
    def delete_update_tags(timestamp, previous, current):
        message = None
        attr_not_clearable = ("name", "unit", "display_name", "display_unit", "data_type", "variable")
        if timestamp:
            
            if len(previous) > len(current): # DELETE TAG

                removed_rows = [row for row in previous if row not in current]
                
                for row in removed_rows:
                    
                    _id = row['id']
                    message = f"Do you want to delete Tag ID: {_id}?"
                    # OPEN MODAL TO CONFIRM CHANGES
                    dash.set_props("modal-update-delete-tag-body", {"children": message})
                    dash.set_props("modal-update_delete-centered", {'is_open': True})

            elif previous and current: # UPDATE TAG DEFINITION
                
                to_updates = find_differences_between_lists(previous, current)
                tag_to_update = to_updates[0]
                tag_id = tag_to_update.pop("id")
                for attr in attr_not_clearable:
                    if attr in tag_to_update:
                        if not tag_to_update[attr]:
                            message = f"You can not empty {attr} attribute"
                    
                if message:
                    dash.set_props("modal-body", {"children": message})
                    dash.set_props("modal-centered", {'is_open': True})
                    return
                message = f"Do you want to update tag {tag_id} To {tag_to_update}?"
                # OPEN MODAL TO CONFIRM CHANGES
                dash.set_props("modal-update-delete-tag-body", {"children": message})
                dash.set_props("modal-update_delete-centered", {'is_open': True})

    @app.callback(
        dash.Output("modal-centered", "is_open"),
        dash.Input("close-centered", "n_clicks"),
        [dash.State("modal-centered", "is_open")],
    )
    def close_error_button(n, is_open):
        r"""
        Documentation here
        """
        if n:

            return not is_open
        
        return is_open
    
    @app.callback(
        dash.Output("modal-success", "is_open"),
        dash.Input("close-success", "n_clicks"),
        [dash.State("modal-success", "is_open")],
    )
    def close_success_button(n, is_open):
        r"""
        Documentation here
        """
        if n:

            return not is_open
        
        return is_open
    
    @app.callback(
        [
            dash.Output("modal-update_delete-centered", "is_open"), 
            dash.Output('tags_datatable', 'data'), 
            dash.Output('tags_datatable', 'data_timestamp'),
            dash.Output("update-delete-tag-yes", "n_clicks"),
            dash.Output("update-delete-tag-no", "n_clicks")
        ],
        [dash.Input("update-delete-tag-yes", "n_clicks"), dash.Input("update-delete-tag-no", "n_clicks")],
        [
            dash.State('tags_datatable', 'data_timestamp'),
            dash.State("modal-update_delete-centered", "is_open"),
            dash.State('tags_datatable', 'data_previous'),
            dash.State('tags_datatable', 'data')
        ]
    )
    def toggle_modal_update_delete_tag(yes_n, no_n, timestamp, is_open, previous, current):
        r"""
        Documentation here
        """
        
        if yes_n:
            
            if timestamp:
                
                if len(previous) > len(current): # DELETE TAG

                    removed_rows = [row for row in previous if row not in current]
                    
                    for row in removed_rows:
                        _id = row['id']
                        message = app.automation.delete_tag(id=_id)
                        
                        if message:
                            dash.set_props("modal-body", {"children": message})
                            dash.set_props("modal-centered", {'is_open': True})
                        
                elif previous and current: # UPDATE TAG DEFINITION
                    to_updates = find_differences_between_lists(previous, current)
                    tag_to_update = to_updates[0]
                    tag_id = tag_to_update.pop("id")
                    if "segment" in tag_to_update:
                        manufacturer_segment = tag_to_update['segment'].split("->")
                        manufacturer = manufacturer_segment[0]
                        segment = manufacturer_segment[1]
                        tag_to_update.update({
                            "segment": segment,
                            "manufacturer": manufacturer
                        })
                    
                    tag, message = app.automation.update_tag(id=tag_id, **tag_to_update)
                    
                    if not tag:
                        dash.set_props("modal-body", {"children": message})
                        dash.set_props("modal-centered", {'is_open': True})

                return not is_open, app.tags_table_data(), None, 0, 0
        
        elif no_n:
            
            return not is_open, app.tags_table_data(), None, 0, 0

        else:

            return is_open, app.tags_table_data(), None, 0, 0
        
    @app.callback(
        [
            dash.Output('alert', 'is_open'),
            dash.Output('alert', 'children'),
            dash.Output('output', 'children')
        ],
        [
            dash.Input('scan_time_input', 'value')
        ],
        [
            dash.State('scan_time_input', 'min'), 
            dash.State('scan_time_input', 'max')
        ]
    )
    def update_scan_time(value, min_value, max_value):
        r"""
        Documentation here
        """
        if value is None:
            
            return False, '', min_value
        
        if value < min_value:

            return True, f'Value {value} is out of range ({min_value}-{max_value})', min_value
        
        if value > max_value:
            
            return True, f'Value {value} is out of range ({min_value}-{max_value})', max_value
        
        return False, '', f'Current value: {value} ms'
            


import dash
import dash_bootstrap_components as dbc
from automation.pages.components import Components
from automation.pages.components.tags import TagsComponents

dash.register_page(__name__)

layout = dbc.Container(
    [
        dbc.Breadcrumb(
            items=[
                {"label": "Home", "href": "/"},
                {"label": "Tags", "active": True},
            ],
        ),
        Components.modal_error(
            title="Error",
            modal_id="modal-centered",
            button_close_id="close-centered",
            body_id="modal-body"
        ),
        Components.modal_error(
            title="Success",
            modal_id="modal-success",
            button_close_id="close-success",
            body_id="modal-success-body"
        ),
        Components.modal_confirm(
            title="Confirmation",
            modal_id="modal-update_delete-centered",
            body_id="modal-update-delete-tag-body",
            yes_button_id="update-delete-tag-yes",
            no_button_id="update-delete-tag-no"
        ),
        TagsComponents.create_tag_form(),
        TagsComponents.tags_table()
    ],
    fluid=False,
    className="my-3"
)


import dash
import dash_bootstrap_components as dbc
from ... import PyAutomation

app = PyAutomation()

class AlarmsComponents:

    @classmethod
    def create_alarm_form(cls):
        r"""
        Documentation here
        """
        return dash.html.Div(
            [
                dash.dcc.Location(id='alarms_page', refresh=False),
                dbc.Accordion(
                    [
                        dbc.AccordionItem(
                            [
                                dbc.Row(
                                    [
                                        dbc.Col(
                                            dbc.InputGroup(
                                                [
                                                    dbc.InputGroupText("Tag"),
                                                    dbc.Select(
                                                        options=[
                                                            {"label": tag["name"], "value": tag["name"]} for tag in app.cvt.get_tags()
                                                        ],
                                                        id="tag_alarm_input"
                                                    ),   
                                                ],
                                                size="md",
                                            ),
                                            width=12,
                                            className="col-sm-12 col-md-3"
                                        ),
                                        
                                        dbc.Col(
                                            dbc.InputGroup(
                                                [
                                                    dbc.Input(placeholder="Alarm Name", id="alarm_name_input"),
                                                ], 
                                                size="md"
                                            ),
                                            width=12,
                                            className="col-sm-12 col-md-3"
                                        ),
                                        
                                        dbc.Col(
                                            dbc.InputGroup(
                                                [
                                                    dbc.InputGroupText(dbc.RadioButton(id="alarm_description_radio_button"), class_name="radiobutton-box"), 
                                                    dbc.Input(placeholder="Alarm Description (Optional)", id="alarm_description_input", disabled=True),
                                                ], 
                                                size="md"
                                            ),
                                            width=12,
                                            className="col-sm-12 col-md-6"
                                        ),

                                        dbc.Col(
                                            dbc.InputGroup(
                                                [
                                                    dbc.InputGroupText("Type"), dbc.Select(options=[], id="alarm_type_input"),
                                                ],
                                                size="md"
                                            ),
                                            width=12,
                                            className="col-sm-12 col-md-3"
                                        ),

                                        dbc.Col(
                                            dbc.InputGroup(
                                                [
                                                    dbc.Input(placeholder="Trigger Value", id="alarm_trigger_value_input"), 
                                                    dbc.InputGroupText('', id="alarm_trigger_unit")
                                                ], 
                                                size="md"
                                            ),
                                            width=12,
                                            className="col-sm-12 col-md-3"
                                        ),

                                        dbc.Col(
                                            dbc.Button(
                                                "Create", 
                                                color="primary", 
                                                outline=True, 
                                                disabled=True, 
                                                id="create_alarm_button",
                                                className="w-100"
                                            ),
                                            width="auto",
                                            className="d-flex justify-content-center align-items-center"
                                        ),
                                    ],
                                    className="form g-3" 
                                ),
                            ],
                            title="Create Alarm",
                            className="my-3"
                        )
                    ],
                    start_collapsed=True,
                )
            ]
        )

    @classmethod
    def alarms_table(cls)->dash.dash_table.DataTable:
        r"""
        Documentation here
        """       
        return dbc.Container(
            dbc.Row(
                dbc.Col(
                    dash.dash_table.DataTable(
                        data=[],
                        columns=[
                            {'name': 'id', 'id': 'id', 'editable': False}, 
                            {'name': 'name', 'id': 'name'}, 
                            {'name': 'tag', 'id': 'tag', 'presentation': 'dropdown'},
                            {'name': 'state', 'id': 'state', 'editable': False},  
                            {'name': 'description', 'id': 'description'}, 
                            {'name': 'alarm_type', 'id': 'alarm_type', 'presentation': 'dropdown'}, 
                            {'name': 'trigger_value', 'id': 'trigger_value', 'type': "numeric"}
                        ],
                        id="alarms_datatable",
                        filter_action="native",
                        sort_action="native",
                        sort_mode="multi",
                        row_deletable=False,
                        selected_columns=[],
                        page_action="native",
                        page_current= 0,
                        page_size= 10,
                        persistence=True,
                        editable=True,
                        persisted_props=['data'],
                        export_format='xlsx',
                        export_headers='display',
                        style_table={'overflowX': 'auto'},
                    ),
                    width=12,
                )
            ),
            fluid=True,
            className="mx-0 px-0"
        )


import dash
from ..pages.components import Components


class ConfigView(dash.Dash):
    r"""
    Documentation here
    """

    def __init__(self, **kwargs):
    
        super(ConfigView, self).__init__(__name__, suppress_callback_exceptions=True, **kwargs)
        
        self.layout = dash.html.Div([
            dash.dcc.Interval(id='timestamp-interval', interval=1000, n_intervals=0),
            Components.navbar(),
            dash.page_container
        ])

    def set_automation_app(self, automation_app):

        self.automation = automation_app
        
    def tags_table_data(self):
    
        tags = list()
        for tag in self.automation.get_tags():
            if tag["segment"] and tag["manufacturer"]:
                tag.update({
                    "segment": f"{tag['manufacturer']}->{tag['segment']}"
                })
            tag.pop('values')
            tag.pop("timestamps")
            tags.append(tag)        
        
        return tags
    
    def alarms_table_data(self):
        data = [{
                "id": alarm["identifier"],
                "tag": alarm["tag"], 
                "name": alarm["name"],
                "description": alarm["description"],
                "state": alarm["state"]["state"],
                "alarm_type": alarm["alarm_setpoint"]["type"],
                "trigger_value": alarm["alarm_setpoint"]["value"],
                } for alarm in self.automation.alarm_manager.serialize()]

        return data
    
    def machines_table_data(self):

        data = [{
                "name": machine["name"],
                "sampling_time": machine["machine_interval"], 
                "description": machine["description"],
                "state": machine["state"],
                "criticity": machine["criticity"],
                "priority": machine["priority"],
                "classification": machine["classification"],
                } for machine in self.automation.machine_manager.serialize_machines()]

        return data


from enum import Enum

class TriggerType(Enum):
    r"""
    Enumeration of alarm trigger types.

    * **HH**: High-High (Critical High)
    * **H**: High
    * **L**: Low
    * **LL**: Low-Low (Critical Low)
    * **B**: Boolean
    * **NONE**: Not Defined
    """
    HH = "HIGH-HIGH"
    H = "HIGH"
    L = "LOW"
    LL = "LOW-LOW"
    B = "BOOL"
    NONE = "NOT DEFINED"


class Trigger:
    r"""
    Represents the condition that triggers an alarm.

    It stores the type of trigger (e.g., HIGH, LOW) and the threshold value.
    """

    def __init__(self):

        self.__type = TriggerType.NONE
        self.__value = None

    @property
    def type(self):
        r"""
        Gets the trigger type.
        """

        return self.__type

    @type.setter
    def type(self, _type:str):
        r"""
        Sets the trigger type.

        **Parameters:**

        * **_type** (str): The trigger type string (e.g., 'HIGH', 'BOOL').
        """
        self.__type = TriggerType(_type)

    @property
    def value(self):
        r"""
        Gets the threshold value.
        """
        return self.__value

    @value.setter
    def value(self, value):
        r"""
        Sets the threshold value.

        Automatically casts the value to `bool` if the trigger type is Boolean,
        otherwise casts to `float` or `int`.

        **Parameters:**

        * **value**: The threshold value.
        """
        if self.type==TriggerType.B:

            if isinstance(value, bool):

                self.__value = value

            else:

                self.__value = bool(value)

        else:

            if isinstance(value, (float, int)):

                self.__value = value

    def serialize(self):
        r"""
        Serializes the trigger object to a dictionary.

        **Returns:**

        * **dict**: {'type': str, 'value': any}
        """
        return {
            "type": self.type.value,
            "value": self.value
        }


# -*- coding: utf-8 -*-
"""automation/logger/logs.py

This module implements the Logs Logger, which tracks application-level logs,
often linking them to specific alarm summaries or events for audit trails.
"""
from datetime import datetime
from ..dbmodels.logs import Logs
from ..modules.users.users import User
from .core import BaseEngine, BaseLogger
from ..utils.decorators import db_rollback


class LogsLogger(BaseLogger):
    r"""
    Logger class specialized for Application Logs.

    This differs from EventsLogger in that it may link logs to specific Alarm occurrences (AlarmSummary) or Events.
    """

    def __init__(self):

        super(LogsLogger, self).__init__()

    @db_rollback
    def create(
        self, 
        message:str, 
        user:User, 
        description:str=None, 
        classification:str=None,
        alarm_summary_id:int=None,
        event_id:int=None,
        timestamp:datetime=None
        ):
        r"""
        Creates a new log entry.

        **Parameters:**

        * **message** (str): Log message.
        * **user** (User): Associated user.
        * **description** (str, optional): Details.
        * **classification** (str, optional): Category.
        * **alarm_summary_id** (int, optional): Link to an alarm history record.
        * **event_id** (int, optional): Link to an event record.
        * **timestamp** (datetime, optional): Time of log.
        """
        if not self.is_history_logged:

            return None
        
        if not self.check_connectivity():

            return list()
            
        query, message = Logs.create(
            message=message, 
            user=user, 
            description=description, 
            classification=classification,
            alarm_summary_id=alarm_summary_id,
            event_id=event_id,
            timestamp=timestamp
        )

        return query, message

    @db_rollback
    def get_lasts(self, lasts:int=1):
        r"""
        Retrieves the last N logs.

        **Parameters:**

        * **lasts** (int): Count to retrieve.
        """
        if not self.is_history_logged:

            return None
        
        if not self.check_connectivity():

            return list()
        
        return Logs.read_lasts(lasts=lasts)
        
    
    @db_rollback
    def filter_by(
        self,
        usernames:list[str]=None,
        alarm_names:list[str]=None,
        event_ids:list[int]=None,
        classification:str="",
        message:str="",
        description:str="",
        greater_than_timestamp:datetime=None,
        less_than_timestamp:datetime=None,
        timezone:str='UTC'
        ):
        r"""
        Filters logs by various criteria.

        **Parameters:**

        * **usernames** (list[str]): Filter by usernames.
        * **alarm_names** (list[str]): Filter by linked alarm names.
        * **event_ids** (list[int]): Filter by linked event IDs.
        * **classification** (str): Filter by classification.
        * **message** (str): Partial match message.
        * **greater_than_timestamp** (datetime): Start time.
        * **less_than_timestamp** (datetime): End time.
        * **timezone** (str): Timezone.
        """
        if not self.is_history_logged:

            return None
        
        if not self.check_connectivity():

            return list()
        
        return Logs.filter_by(
            usernames=usernames,
            alarm_names=alarm_names,
            event_ids=event_ids,
            message=message,
            description=description,
            classification=classification,
            greater_than_timestamp=greater_than_timestamp,
            less_than_timestamp=less_than_timestamp,
            timezone=timezone
        )

    @db_rollback  
    def get_summary(self)->tuple[list, str]:
        r"""
        Retrieves a summary of all logs.
        """
        if not self.is_history_logged:

            return None
        
        if not self.check_connectivity():
            
            return list()
            
        return Logs.serialize()
    
    
class LogsLoggerEngine(BaseEngine):
    r"""
    Thread-safe Engine for the LogsLogger.
    """

    def __init__(self):

        super(LogsLoggerEngine, self).__init__()
        self.logger = LogsLogger()

    def create(
        self,
        message:str, 
        user:User, 
        description:str=None, 
        classification:str=None,
        alarm_summary_id:int=None,
        event_id:int=None,
        timestamp:datetime=None
        ):
        r"""
        Thread-safe log creation.
        """
        _query = dict()
        _query["action"] = "create"
        _query["parameters"] = dict()
        _query["parameters"]["message"] = message
        _query["parameters"]["user"] = user
        _query["parameters"]["description"] = description
        _query["parameters"]["classification"] = classification
        _query["parameters"]["alarm_summary_id"] = alarm_summary_id
        _query["parameters"]["event_id"] = event_id
        _query["parameters"]["timestamp"] = timestamp
        
        return self.query(_query)
    
    def get_lasts(
        self,
        lasts:int=1
        ):
        r"""
        Thread-safe retrieval of last logs.
        """
        _query = dict()
        _query["action"] = "get_lasts"
        _query["parameters"] = dict()
        _query["parameters"]["lasts"] = lasts
        
        return self.query(_query)
    
    def filter_by(
        self,
        usernames:list[str]=None,
        alarm_names:list[str]=None,
        event_ids:list[int]=None,
        classification:str="",
        message:str="",
        description:str="",
        greater_than_timestamp:datetime=None,
        less_than_timestamp:datetime=None,
        timezone:str='UTC'
        ):
        r"""
        Thread-safe log filtering.
        """
        _query = dict()
        _query["action"] = "filter_by"
        _query["parameters"] = dict()
        _query["parameters"]["usernames"] = usernames
        _query["parameters"]["alarm_names"] = alarm_names
        _query["parameters"]["event_ids"] = event_ids
        _query["parameters"]["classification"] = classification
        _query["parameters"]["message"] = message
        _query["parameters"]["description"] = description
        _query["parameters"]["greater_than_timestamp"] = greater_than_timestamp
        _query["parameters"]["less_than_timestamp"] = less_than_timestamp
        _query["parameters"]["timezone"] = timezone
        
        return self.query(_query)

    def get_summary(self):
        r"""
        Thread-safe retrieval of log summary.
        """
        _query = dict()
        _query["action"] = "get_summary"
        _query["parameters"] = dict()
        
        return self.query(_query)



from ..utils.units import *

class MassFlow(EngUnit):
    """Creates a mass flow object that can store a mass flow value and 
    convert between units of mass flow.

    :param value: [int|float] Engineering value\n
    :param unit: [str] Engineering unit\n
    :return: [MassFlow Object]\n

    ```python
    >>> from automation.variables.mass_flow import MassFlow
    >>> mass_flow = MassFlow(value=1.0, unit="kg/hr")
    >>> mass_flow.value
    1.0
    >>> mass_flow.unit
    'kg/hr'
    >>> mass_flow.convert(to_unit="lb/sec")
    0.0006123951727222222
    >>> MassFlow.convert_value(value=1.0, from_unit="kg/hr", to_unit="lb/sec")
    0.0006123951727222222
    >>> MassFlow.convert_values(values=[1.0, 10.0], from_unit="kg/hr", to_unit="lb/sec")
    [0.0006123951727222222, 0.0061239517272222216]
    >>> mass_flow.change_unit(unit="lb/sec")
    0.0006123951727222222
    >>> mass_flow.unit
    'lb/sec'
    >>> mass_flow.get_value()
    [0.0006123951727222222, 'lb/sec']
    >>> print(mass_flow)
    0.0006123951727222222 lb/sec
    >>> mass_flow2 = MassFlow(value=3.0, unit="lb/sec")
    >>> mass_flow_result = mass_flow + mass_flow2
    >>> print(mass_flow_result)
    3.0006123951727224 lb/sec
    >>> mass_flow_result = mass_flow * mass_flow2
    >>> print(mass_flow_result)
    0.0018371855181666666 lb/sec
    >>> mass_flow_result = mass_flow / mass_flow2
    >>> print(mass_flow_result)
    0.00020413172424074073 lb/sec
    >>> mass_flow_result = mass_flow ** mass_flow2
    >>> print(mass_flow_result)
    2.2966524349040475e-10 lb/sec

    ```

    """
    
    class Units(UnitSerializer):
        kg_day = 'kg/day'
        kg_hr = 'kg/hr'
        kg_min = 'kg/min'
        kg_sec = 'kg/sec'
        g_day = 'g/day'
        g_hr = 'g/hr'
        g_min = 'g/min'
        g_sec = 'g/sec'
        mg_day = 'mg/day'
        mg_hr = 'mg/hr'
        mg_min = 'mg/min'
        mg_sec = 'mg/sec'
        lb_day = 'lb/day'
        lb_hr = 'lb/hr'
        lb_min = 'lb/min'
        lb_sec = 'lb/sec'
        metricTon_day = 'metricTon/day'
        metricTon_hr = 'metricTon/hr'
        metricTon_min = 'metricTon/min'
        metricTon_sec = 'metricTon/sec'
    
    conversions = {
        'kg/day' : 1.0,
        'kg/hr' : 1.0 / 24,
        'kg/min' : 1.0 / 24 / 60,
        'kg/sec' : 1.0 / 24 / 60 / 60,
        'g/day' : 1000.0,
        'g/hr' : 1000.0 / 24,
        'g/min' : 1000.0 / 24 / 60,
        'g/sec' : 1000.0 / 24 / 60 / 60,
        'mg/day' : 1000000.0,
        'mg/hr' : 1000000.0 / 24,
        'mg/min' : 1000000.0 / 24 / 60,
        'mg/min' : 1000000.0 / 24 / 60 / 60,
        'metricTon/day' : 1.0 / 1000.0,
        'metricTon/hr' : 1.0 / 1000.0 / 24,
        'metricTon/min' : 1.0 / 1000.0 / 24 / 60,
        'metricTon/sec' : 1.0 / 1000.0 / 24 / 60 / 60,
        'lb/day' : 2.2046226218,
        'lb/hr' : 2.2046226218 / 24,
        'lb/min' : 2.2046226218 / 24 / 60,
        'lb/sec' : 2.2046226218 / 24 / 60 / 60
    }

    def __init__(self, value, unit):

        if unit not in MassFlow.Units.list():

            raise UnitError(f"{unit} value is not allowed for {self.__class__.__name__} object - you can use: {MassFlow.Units.list()}")
        
        super(MassFlow, self).__init__(value=value, unit=unit)


import pytz
from peewee import CharField, FloatField, ForeignKeyField, TimestampField
from ..dbmodels.core import BaseModel 
from datetime import datetime
from .tags import Tags
from ..alarms.states import States
from ..tags.cvt import CVTEngine

tag_engine = CVTEngine()

class AlarmTypes(BaseModel):
    r"""
    Database model for Alarm Types (e.g., HIGH, LOW, BOOL).
    """

    name = CharField(unique=True)                       # high-high , high , bool , low , low-low

    @classmethod
    def create(cls, name:str)-> dict:
        r"""
        Creates a new Alarm Type if it doesn't exist.

        **Parameters:**

        * **name** (str): The name of the alarm type (e.g., 'HIGH').

        **Returns:**

        * **dict**: Result dictionary containing status message and data.
        """
        result = dict()
        data = dict()
        name = name.upper()

        if not cls.name_exist(name):

            query = cls(name=name)
            query.save()
            
            message = f"Alarm type {name} created successfully"
            data.update(query.serialize())

            result.update(
                {
                    'message': message, 
                    'data': data
                }
            )
            return result

        message = f"Alarm type {name} is already into database"
        result.update(
            {
                'message': message, 
                'data': data
            }
        )
        return result

    @classmethod
    def read_by_name(cls, name:str):
        r"""
        Retrieves an Alarm Type by name.

        **Parameters:**

        * **name** (str): Alarm type name.

        **Returns:**

        * **AlarmTypes**: The model instance or None.
        """        
        return cls.get_or_none(name=name.upper())

    @classmethod
    def name_exist(cls, name:str)->bool:
        r"""
        Checks if an Alarm Type name exists.

        **Parameters:**

        * **name** (str): Alarm type name.

        **Returns:**

        * **bool**: True if exists.
        """
        query = cls.get_or_none(name=name.upper())
        
        if query is not None:

            return True
        
        return False

    def serialize(self)-> dict:
        r"""
        Serializes the record to a dictionary.
        """

        return {
            "id": self.id,
            "name": self.name
        }


class AlarmStates(BaseModel):
    r"""
    Database model for Alarm States (ISA 18.2).
    """
    name = CharField(unique=True)
    mnemonic = CharField(max_length=20)
    condition = CharField(max_length=20)
    status = CharField(max_length=20)

    @classmethod
    def create(cls, name:str, mnemonic:str, condition:str, status:str)-> dict:
        r"""
        Creates a new Alarm State.

        **Parameters:**

        * **name** (str): State name (e.g., 'Unacknowledged').
        * **mnemonic** (str): Short code (e.g., 'UNACK').
        * **condition** (str): Process condition.
        * **status** (str): Status description.

        **Returns:**

        * **AlarmStates**: The created instance or existing one.
        """

        if not cls.name_exist(name):

            query = cls(name=name, mnemonic=mnemonic, condition=condition, status=status)
            query.save()
            
            return query

    @classmethod
    def read_by_name(cls, name:str):
        r"""
        Retrieves an Alarm State by name.
        """
        return cls.get_or_none(name=name)

    @classmethod
    def name_exist(cls, name:str)->bool:
        r"""
        Checks if an Alarm State name exists.
        """
        query = cls.get_or_none(name=name)
        
        if query is not None:

            return True
        
        return False

    def serialize(self)-> dict:
        r"""
        Serializes the record.
        """

        return {
            "id": self.id,
            "name": self.name,
            "mnemonic": self.mnemonic,
            "condition": self.condition,
            "status": self.status
        }


class Alarms(BaseModel):
    r"""
    Database model for configured Alarms.
    """
    
    identifier = CharField(unique=True)
    name = CharField(unique=True, max_length=128)
    tag = ForeignKeyField(Tags, backref='alarms')
    trigger_type = ForeignKeyField(AlarmTypes, backref='alarms')
    trigger_value = FloatField()
    description = CharField(null=True, max_length=256)
    state = ForeignKeyField(AlarmStates, backref='alarms')
    timestamp = TimestampField(utc=True, null=True)

    @classmethod
    def create(
        cls,
        identifier:str,
        name:str,
        tag:str,  
        trigger_type:str,
        trigger_value:float,
        description:str=None,
        state:str=States.NORM.value,
        timestamp:datetime=None
        ):
        r"""
        Creates a new Alarm configuration record.

        **Parameters:**

        * **identifier** (str): Unique ID.
        * **name** (str): Alarm name.
        * **tag** (str): Associated Tag name.
        * **trigger_type** (str): Type of trigger.
        * **trigger_value** (float): Threshold value.
        * **description** (str, optional): Description.
        * **state** (str, optional): Initial state.
        * **timestamp** (datetime, optional): Creation timestamp.

        **Returns:**

        * **Alarms**: The created alarm record.
        """
        trigger_type = AlarmTypes.read_by_name(name=trigger_type)
        tag = Tags.read_by_name(name=tag)
        state = AlarmStates.read_by_name(name=state)
        if not cls.name_exists(name):

            alarm = super().create(
                identifier=identifier,
                name=name,
                tag=tag,  
                trigger_type=trigger_type,
                trigger_value=trigger_value,
                description=description,
                state=state,
                timestamp=timestamp
            )
            alarm.save()

            return alarm
    
    @classmethod
    def name_exists(cls, name:str)->bool|None:
        r"""
        Checks if an alarm name exists.
        """
        tag = cls.get_or_none(name=name)
        if tag:

            return True
        
    @classmethod
    def read(cls, id:str):
        r"""
        Reads an alarm by ID.
        """
        return cls.get_or_none(id=id)
    
    @classmethod
    def read_by_identifier(cls, identifier:str):
        r"""
        Reads an alarm by unique identifier.
        """
        return cls.get_or_none(identifier=identifier)
        
    @classmethod
    def read_by_name(cls, name:str):
        r"""
        Reads an alarm by name.
        """
        return cls.get_or_none(name=name)

    def serialize(self):
        r"""
        Serializes the alarm record.
        """
        timestamp = self.timestamp
        if timestamp:

            timestamp = timestamp.strftime(tag_engine.DATETIME_FORMAT)

        return {
            'identifier': self.identifier,
            'name': self.name,
            'tag': self.tag.name,  
            'alarm_type': self.trigger_type.name,
            'trigger_value': self.trigger_value,
            'description': self.description,
            'state': self.state.name,
            'timestamp': timestamp
        }
    

class AlarmSummary(BaseModel):
    r"""
    Database model for Alarm History (Summary).
    """
    
    alarm = ForeignKeyField(Alarms, backref='summary')
    state = ForeignKeyField(AlarmStates, backref='summary')
    alarm_time = TimestampField(utc=True)
    ack_time = TimestampField(utc=True, null=True)

    @classmethod
    def create(cls, name:str, state:str, timestamp:datetime, ack_timestamp:datetime=None):
        r"""
        Creates a new entry in the alarm summary.

        **Parameters:**

        * **name** (str): Alarm name.
        * **state** (str): Alarm state.
        * **timestamp** (datetime): Time of occurrence.
        * **ack_timestamp** (datetime, optional): Acknowledgment time.
        """
        _alarm = Alarms.read_by_name(name=name)
        _state = AlarmStates.read_by_name(name=state)
        
        if _alarm:

            if _state:

                # Create record
                query = cls(alarm=_alarm.id, state=_state.id, alarm_time=timestamp, ack_time=ack_timestamp)
                query.save()
                
                return query

    @classmethod
    def read_by_name(cls, name:str):
        r"""
        Retrieves the latest summary entry for a specific alarm name.
        """
        alarm = Alarms.read_by_name(name=name)
        return cls.select().where(cls.alarm==alarm).order_by(cls.id.desc()).get_or_none()

    @classmethod
    def read_by_alarm_id(cls, alarm_id:int):
        r"""
        Retrieves the latest summary entry by alarm ID.
        """
        alarm = Alarms.read(id=alarm_id)
        return cls.select().where(cls.alarm==alarm).order_by(cls.id.desc()).get_or_none()

    @classmethod
    def read_all(cls):
        r"""
        Retrieves all alarm summary records.
        """
        data = list()
        
        try:
            data = [query.serialize() for query in cls.select().order_by(cls.id.desc())]

            return data

        except Exception as _err:

            return data

    @classmethod
    def read_lasts(cls, lasts:int=1):
        r"""
        Retrieves the last N records.
        """
        alarms = cls.select().order_by(cls.id.desc()).limit(lasts)

        return [alarm.serialize() for alarm in alarms]
    
    @classmethod
    def filter_by(
        cls,
        states:list[str]=None,
        names:list[str]=None,
        tags:list[str]=None,
        greater_than_timestamp:datetime=None,
        less_than_timestamp:datetime=None,
        timezone:str='UTC',
        page:int=1,
        limit:int=20
        ):
        r"""
        Filters alarm summary records with pagination.

        **Parameters:**

        * **states** (list[str]): Filter by states.
        * **names** (list[str]): Filter by alarm names.
        * **tags** (list[str]): Filter by tag names.
        * **greater_than_timestamp** (datetime): Start time.
        * **less_than_timestamp** (datetime): End time.
        * **page**, **limit**: Pagination control.

        **Returns:**

        * **dict**: {data: list, pagination: dict}
        """
        import math
        _timezone = pytz.timezone(timezone)
        query = cls.select()
        
        if states:
            subquery = AlarmStates.select(AlarmStates.id).where(AlarmStates.name.in_(states))
            query = query.join(AlarmStates).where(AlarmStates.id.in_(subquery))
        
        if names:
            subquery = Alarms.select(Alarms.id).where(Alarms.name.in_(names))
            query = query.join(Alarms).where(Alarms.id.in_(subquery))
        
        if tags:
            subquery = Tags.select(Tags.id).where(Tags.name.in_(tags))
            subquery = Alarms.select(Alarms.id).join(Tags).where(Tags.id.in_(subquery))
            query = query.join(Alarms).where(Alarms.id.in_(subquery))
        
        if greater_than_timestamp:
            greater_than_timestamp = _timezone.localize(datetime.strptime(greater_than_timestamp, '%Y-%m-%d %H:%M:%S.%f')).astimezone(pytz.UTC)
            query = query.where(cls.alarm_time > greater_than_timestamp)
        
        if less_than_timestamp:
            less_than_timestamp = _timezone.localize(datetime.strptime(less_than_timestamp, '%Y-%m-%d %H:%M:%S.%f')).astimezone(pytz.UTC)
            query = query.where(cls.alarm_time < less_than_timestamp)
        
        query = query.order_by(cls.id.desc())
        
        total_records = query.count()
        
        if limit <= 0: limit = 20
        if page <= 0: page = 1
        
        total_pages = math.ceil(total_records / limit)
        if total_pages == 0: total_pages = 1
        
        has_next = page < total_pages
        has_prev = page > 1
        
        paginated_query = query.paginate(page, limit)
        
        data = [alarm.serialize() for alarm in paginated_query]
        
        return {
            "data": data,
            "pagination": {
                "page": page,
                "limit": limit,
                "total_records": total_records,
                "total_pages": total_pages,
                "has_next": has_next,
                "has_prev": has_prev
            }
        }

    @classmethod
    def get_alarm_summary_comments(cls, id:int):
        r"""
        Retrieves comments associated with a specific alarm summary entry.
        """
        query = cls.read(id=id)

        return [comment.serialize() for comment in query.logs]

    def serialize(self):
        r"""
        Serializes the summary record.
        """
        from .. import TIMEZONE

        ack_time = None
        if self.ack_time:
            ack_time = self.ack_time
            ack_time = pytz.UTC.localize(ack_time).astimezone(TIMEZONE)
            ack_time = ack_time.strftime(tag_engine.DATETIME_FORMAT)

        alarm_time = self.alarm_time
        alarm_time = pytz.UTC.localize(alarm_time).astimezone(TIMEZONE)
        alarm_time = alarm_time.strftime(tag_engine.DATETIME_FORMAT)

        return {
            'id': self.id,
            'name': self.alarm.name,
            'tag': self.alarm.tag.name,
            'description': self.alarm.description,
            'state': self.state.name,
            'mnemonic': self.state.mnemonic,
            'status': self.state.status,
            'alarm_time': alarm_time,
            'ack_time': ack_time,
            'has_comments': True if self.logs else False
        }
    


# -*- coding: utf-8 -*-
"""automation/workers/state.py

This module implements the State Machine Worker, managing the execution of state machines.
"""
import heapq
import logging
import time
from collections import deque
from threading import Thread
from .worker import BaseWorker


class MachineScheduler():
    r"""
    A simple scheduler for executing tasks (state machine loops) periodically.

    It maintains a queue of ready tasks and a heap of scheduled tasks.
    """

    def __init__(self):

        self._ready = deque()
        self._sleeping = list()
        self._sequence = 0
        self.last = None
        self._stop = False

    def call_soon(self, func):
        r"""
        Schedules a function to be called as soon as possible.

        **Parameters:**

        * **func** (callable): The function to execute.
        """
        self._ready.append(func)

    def call_later(self, delay, func, machine):
        r"""
        Schedules a function to be called after a delay.

        **Parameters:**

        * **delay** (float): Delay in seconds.
        * **func** (callable): The function to execute.
        * **machine** (StateMachine): The associated state machine instance.
        """
        self._sequence += 1
        deadline = time.time() + delay
        heapq.heappush(self._sleeping, (deadline, self._sequence, func, machine))

    def stop(self):
        r"""
        Stops the scheduler loop.
        """
        self._stop = True
    
    def run(self):
        r"""
        Main scheduler loop.

        Executes tasks in the ready queue and moves scheduled tasks to the ready queue
        when their deadline is reached. Handles sleep intervals to manage CPU usage.
        """
        self.set_last()
        
        while self._ready or self._sleeping:

            if self._stop:
                break

            if not self._ready and self._sleeping:
                deadline, _, func, machine = heapq.heappop(self._sleeping)
                self.sleep_elapsed(machine)
                
                self._ready.append(func)

            while self._ready:
                func = self._ready.popleft()
                func()

    def set_last(self):
        r"""
        Updates the last execution timestamp.
        """
        self.last = time.time()

        return self.last

    def sleep_elapsed(self, machine):
        r"""
        Sleeps for the remaining time until the next scheduled task.

        **Parameters:**

        * **machine** (StateMachine): The machine associated with the next task.
        """
        elapsed = time.time() - self.last
        interval = machine.get_interval()
        
        try:
            time.sleep(interval - elapsed)
            self.set_last()
        except ValueError:
            self.set_last()
            logger = logging.getLogger("pyautomation")
            logger.warning(f"State Machine: {machine.name.value} NOT executed on time - Execution Interval: {interval} - Elapsed: {elapsed}")


class SchedThread(Thread):
    r"""
    A thread that runs a dedicated scheduler for a single state machine.
    """

    def __init__(self, machine):

        super(SchedThread, self).__init__()

        self.machine = machine

    def stop(self):
        r"""
        Stops the scheduler running in this thread.
        """
        self.scheduler.stop()

    def loop_closure(self, machine, scheduler:MachineScheduler):
        r"""
        Creates a closure for the state machine loop function.

        **Parameters:**

        * **machine** (StateMachine): The state machine.
        * **scheduler** (MachineScheduler): The scheduler managing execution.

        **Returns:**

        * **callable**: The loop function.
        """
        def loop():
            machine.loop()
            interval = machine.get_interval()
            scheduler.call_later(interval, loop, machine)
    
        return loop
    
    def target(self, machine):
        r"""
        The target function for the thread. Initializes and runs the scheduler.
        """
        scheduler = MachineScheduler()
        self.scheduler = scheduler
        func = self.loop_closure(machine, scheduler)
        scheduler.call_soon(func)
        scheduler.run() 

    def run(self):
        r"""
        Starts the thread execution.
        """
        self.target(self.machine)


class AsyncStateMachineWorker(BaseWorker):
    r"""
    Worker that manages asynchronously executed state machines (each in its own thread).
    """

    def __init__(self):

        super(AsyncStateMachineWorker, self).__init__()
        self._machines = list()
        self._schedulers = list()
        self.jobs = list()

    def add_machine(self, machine):
        r"""
        Adds a machine to be managed by this worker.
        """
        self._machines.append(machine)

    def run(self):
        r"""
        Starts a separate thread (SchedThread) for each registered machine.
        """
        for machine in self._machines:

            sched = SchedThread(machine)
            self._schedulers.append(sched)

        for sched in self._schedulers:

            sched.daemon = True
            sched.start()

    def join(self, machine):
        r"""
        Adds and starts a new machine dynamically at runtime.
        """
        sched = SchedThread(machine)
        self._schedulers.append(sched)
        sched.daemon = True
        sched.start()

    def drop(self, machine):
        r"""
        Stops and removes a machine from execution.
        """
        sched_to_drop = None
        for index, sched in enumerate(self._schedulers):
            if machine==sched.machine:

                sched_to_drop = self._schedulers.pop(index)
                break
        
        if sched_to_drop:
            sched.stop()

    def stop(self):
        r"""
        Stops all managed threads.
        """
        for sched in self._schedulers:
            try:
                sched.stop()
            except Exception as e:
                message = "Error on async scheduler stop"
                logger = logging.getLogger("pyautomation")
                logger.error(f"{message} - {e}")
    

class StateMachineWorker(BaseWorker):
    r"""
    The main worker responsible for coordinating state machine execution.

    It manages two types of execution:
    1. **Sync**: Machines executed sequentially in the main worker thread (cooperative multitasking).
    2. **Async**: Machines executed in separate threads (preemptive multitasking).
    """

    def __init__(self, manager):

        super(StateMachineWorker, self).__init__()
        
        self._manager = manager
        self._sync_scheduler = MachineScheduler()
        self._async_scheduler = AsyncStateMachineWorker()
        self.jobs = list()

    def loop_closure(self, machine):
        
        self._machine = machine

        def loop():
            machine.loop()
            interval = machine.get_interval()
            self._sync_scheduler.call_later(interval, loop, machine)

        return loop

    def run(self):
        r"""
        Starts the worker.

        Iterates through registered machines and assigns them to either the sync or async scheduler
        based on their configuration.
        """
        for machine, interval, mode in self._manager.get_machines():
    
            if mode == "async":
                
                self._async_scheduler.add_machine(machine)                
                
            else:

                func = self.loop_closure(machine)
                self._sync_scheduler.call_soon(func)

        self._async_scheduler.run()
        self._sync_scheduler.run()

    def stop(self):
        r"""
        Stops both sync and async schedulers.
        """
        self._async_scheduler.stop()
        self._sync_scheduler.stop()



import pytz
from datetime import datetime, timedelta
from flask_restx import Namespace, Resource, fields
from .... import PyAutomation
from ....extensions.api import api
from ....extensions import _api as Api
from ....dbmodels.events import Events
from .... import _TIMEZONE, TIMEZONE

ns = Namespace('Events Logger', description='System Event Logging and Retrieval')
app = PyAutomation()


events_filter_model = api.model("events_filter_model",{
    'usernames': fields.List(fields.String(), required=False, description='List of usernames to filter by'),
    'priorities': fields.List(fields.Integer(), required=False, description='List of priority levels to filter by'),
    'criticities': fields.List(fields.Integer(), required=False, description='List of criticality levels to filter by'),
    'message': fields.String(required=False, description='Partial message content to search for'),
    'classification': fields.String(required=False, description='Event classification/category'),
    'description': fields.String(required=False, description='Partial description content to search for'),
    'greater_than_timestamp': fields.DateTime(required=False, default=datetime.now(pytz.utc).astimezone(TIMEZONE) - timedelta(minutes=30), description=f'Start time for filtering - DateTime Format: {app.cvt.DATETIME_FORMAT}'),
    'less_than_timestamp': fields.DateTime(required=False, default=datetime.now(pytz.utc).astimezone(TIMEZONE), description=f'End time for filtering - DateTime Format: {app.cvt.DATETIME_FORMAT}'),
    'timezone': fields.String(required=False, default=_TIMEZONE, description='Timezone for the query'),
    'page': fields.Integer(required=False, default=1, description='Page number for pagination'),
    'limit': fields.Integer(required=False, default=20, description='Items per page')
})

    
@ns.route('/filter_by')
class EventsSummaryFilterByResource(Resource):

    @api.doc(security='apikey', description="Filters system events based on criteria.")
    @api.response(200, "Success")
    @api.response(400, "Invalid parameters")
    @Api.token_required(auth=True)
    @ns.expect(events_filter_model)
    def post(self):
        r"""
        Filter system events.

        Retrieves a list of system events matching the provided filters.
        """
        timezone = _TIMEZONE
        if "timezone" in api.payload:

            timezone = api.payload["timezone"]

        if timezone not in pytz.all_timezones:

            return f"Invalid Timezone", 400
        
        separator = '.'
        if 'greater_than_timestamp' in api.payload:
            
            greater_than_timestamp = api.payload['greater_than_timestamp']
            api.payload['greater_than_timestamp'] = greater_than_timestamp.replace("T", " ").split(separator, 1)[0] + '.00'
        
        if "less_than_timestamp" in api.payload:

            less_than_timestamp = api.payload['less_than_timestamp']
            api.payload['less_than_timestamp'] = less_than_timestamp.replace("T", " ").split(separator, 1)[0] + '.00'
        
        return app.filter_events_by(**api.payload)
    

@ns.route('/lasts/<lasts>')
@api.param('lasts', 'Number of records to retrieve')
class LastsEventsResource(Resource):

    @api.doc(security='apikey', description="Retrieves the last N system events.")
    @api.response(200, "Success")
    @Api.token_required(auth=True)
    def get(self, lasts:int=10):
        r"""
        Get latest events.

        Retrieves the most recent system events from the log.
        """
        
        return app.get_lasts_events(lasts=int(lasts))
    

@ns.route('/<id>/comments')
@api.param('id', 'Event ID')
class EventsCommentsResource(Resource):

    @api.doc(security='apikey', description="Retrieves comments associated with a specific event.")
    @api.response(200, "Success")
    @Api.token_required(auth=True)
    def get(self, id:int):
        r"""
        Get event comments.

        Retrieves user comments linked to a specific system event.
        """
        comments = Events.get_comments(id=id)
        return comments, 200



# -*- coding: utf-8 -*-
"""automation/managers/state_machine.py

This module implements the State Machine Manager, which acts as a registry and controller for all
state machine instances within the application.
"""
from statemachine import StateMachine
from ..models import StringType
from ..tags import TagObserver, CVTEngine, Tag
import queue

class StateMachineManager:
    r"""
    Manages the lifecycle and execution configuration of State Machines.

    It maintains a registry of all active machines, their execution intervals, and threading modes (sync/async).
    """

    def __init__(self):

        self._machines = list()
        self._tag_queue = queue.Queue()

    def get_queue(self)->queue.Queue:
        r"""
        Retrieves the internal queue used for tag updates related to state machines.
        """
        return self._tag_queue

    def append_machine(self, machine:StateMachine):
        r"""
        Registers a new state machine.

        **Parameters:**

        * **machine** (StateMachine): The state machine instance.
        """
        
        self._machines.append(machine)

    def get_machines(self)->list:
        r"""
        Retrieves the list of all registered state machines.

        **Returns:**

        * **list**: List of tuples [(machine, interval, mode), ...].
        """
        result = self._machines
        
        return result
    
    def serialize_machines(self):
        r"""
        Serializes all registered machines to a list of dictionaries.

        **Returns:**

        * **list[dict]**: Serialized machine data.
        """

        return [machine.serialize() for machine, _, _ in self.get_machines()]

    def get_machine(self, name:StringType)->StateMachine:
        r"""
        Retrieves a state machine by its name.

        **Parameters:**

        * **name** (StringType): The name of the machine.

        **Returns:**

        * **StateMachine**: The machine instance if found.
        """
        for machine, _, _ in self._machines:

            if name.value == machine.name.value:

                return machine
            
    def drop(self, name:str):
        r"""
        Removes a state machine from the manager.

        **Parameters:**

        * **name** (str): The name of the machine to remove.

        **Returns:**

        * **tuple**: The removed (machine, interval, mode) tuple.
        """
        index = 0
        for machine, _, _ in self._machines:

            if name == machine.name.value:

                machine_to_revome_from_worker = self._machines.pop(index)
                break

            index += 1

        if machine_to_revome_from_worker:

            return machine_to_revome_from_worker

    def unsubscribe_tag(self, tag:Tag):
        r"""
        Unsubscribes a tag from all state machines. 
        
        If a DAQ machine has no more subscribed tags, it is removed.

        **Parameters:**

        * **tag** (Tag): The tag to unsubscribe.

        **Returns:**

        * **tuple**: The removed machine tuple if a DAQ machine was dropped.
        """
        machine_to_revome_from_worker = (None, None, None)
        for machine, _, _ in self._machines:

            if hasattr(machine, "unsubscribe_to"):

                machine.unsubscribe_to(tag=tag)

                if machine.classification.value.lower()=="data acquisition system":

                    if not machine.get_subscribed_tags():
                
                        machine_to_revome_from_worker = self.drop(name=machine.name.value)
                        break

        if machine_to_revome_from_worker:

            return machine_to_revome_from_worker

    def summary(self)->dict:
        r"""
        Generates a summary of registered state machines.

        **Returns:**

        * **dict**: {length: int, state_machines: list[str]}
        """
        result = dict()
        machines = [machine.name for machine, _, _ in self.get_machines()]

        result["length"] = len(machines)
        result["state_machines"] = machines

        return result

    def exist_machines(self)->bool:
        r"""
        Checks if there are any registered state machines.

        **Returns:**

        * **bool**: True if at least one machine exists.
        """
        return len(self._machines) > 0


from ..utils.units import *

class Force(EngUnit):
    """Creates a force object that can store a force value and 
    convert between units of force.

    :param value: [int|float] Engineering value\n
    :param unit: [str] Engineering unit\n
    :return: [Force Object]\n

    ```python
    >>> from automation.variables.force import Force
    >>> force = Force(value=1.0, unit="N")
    >>> force.value
    1.0
    >>> force.unit
    'N'
    >>> force.convert(to_unit="kgf")
    0.1019716213
    >>> Force.convert_value(value=1.0, from_unit="N", to_unit="lbf")
    0.2248089431
    >>> Force.convert_values(values=[1.0, 10.0], from_unit="N", to_unit="lbf")
    [0.2248089431, 2.248089431]
    >>> force.change_unit(unit="kgf")
    0.1019716213
    >>> force.unit
    'kgf'
    >>> force.get_value()
    [0.1019716213, 'kgf']
    >>> print(force)
    0.1019716213 kgf
    >>> force2 = Force(value=3.0, unit="lbf")
    >>> force_result = force + force2
    >>> print(force_result)
    1.4627487313277014 kgf
    >>> force_result = force * force2
    >>> print(force_result)
    0.1387606481374532 kgf
    >>> force_result = force / force2
    >>> print(force_result)
    0.07493631436666666 kgf
    >>> force_result = force ** force2
    >>> print(force_result)
    0.04474673588461198 kgf

    ```
    """

    class Units(UnitSerializer):
        N = 'N'
        kN = 'kN'
        MN = 'MN'
        GN = 'GN'
        gf = 'gf'
        kgf = 'kgf'
        dyn = 'dyn'
        Jm = 'J/m'
        Jcm = 'J/cm'
        shortTonF = 'shortTonF'
        longTonF = 'longTonF'
        kipf = 'kipf'
        lbf = 'lbf'
        ozf = 'ozf'
        pdl = 'pdl'

    conversions = {
        'N' : 1.0,
        'kN' : 1.0 / 1000.0,
        'MN' : 1.0 / 1000000.0,
        'GN' : 1.0 / 1000000000.0,
        'gf' : 1.019716213e+2,
        'kgf' : 1.019716213e-1,
        'dyn' : 1e+5,
        'J/m' : 1.0,
        'J/cm' : 100.0,
        'shortTonF' : 1.124045e-4,
        'longTonF' : 1.003611e-4,
        'kipf' : 2.248089e-4,
        'lbf' : 2.248089431e-1,
        'ozf' : 3.5969430896,
        'pdf' : 7.2330138512
    }

    def __init__(self, value, unit):

        if unit not in Force.Units.list():

            raise UnitError(f"{unit} value is not allowed for {self.__class__.__name__} object - you can use: {Force.Units.list()}")
        
        super(Force, self).__init__(value=value, unit=unit)